/*
* This Java Bean and all of the other "Beans" in this offering are
* copyright 2002 by Peter Wayner. 
*   


 * This bean is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This bean is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * Or go to http://www.gnu.org
 * 
 
 */
import java.util.Vector;
import java.io.*;
import java.math.BigInteger;

class StringConverterNode{
  String c;
  int value;
  int bits;
  String bitString;
}


public class StringConverter{
   // converts Strings into numbers using huffman compression.
   // This code could be more efficient with array lookup instead
   // of findNode, but I want to make it easy to expand to bigger
   // dimensions.
   
   
   // The array basicEnglish is hard coded to avoid loading a separate file.
   // In the future, I hope to use digraphs to do a better job compressing
   // the message.
   
   /// Note, there's a bit of cute engineering going on here. The bits encoded
   /// into the message are padded with 0s. If there are too many zeros coming
   //// out, they'll be turned into spaces. Three zeros at a time. Padding into
   //// padding. 
   
/* String[] basicEnglish={"bitString=0000 c=T count=30539 bits=4 value=0",
"bitString=00010 c=H count=14892 bits=5 value=2",
"bitString=000110 c=F count=7763 bits=6 value=6",
"bitString=000111 c=P count=6977 bits=6 value=7",
"bitString=001 c=  count=59159 bits=3 value=1",
"bitString=010000 c=B count=6642 bits=6 value=16",
"bitString=010001 c=G count=6527 bits=6 value=17",
"bitString=01001000 c=K count=1841 bits=8 value=72",
"bitString=010010010000 c=6 count=123 bits=12 value=1168",
"bitString=010010010001 c=? count=123 bits=12 value=1169",
"bitString=010010010010 c== count=118 bits=12 value=1170",
"bitString=0100100100110 c=7 count=68 bits=13 value=2342",
"bitString=01001001001110 c=; count=26 bits=14 value=4686",
"bitString=010010010011110 c=< count=12 bits=15 value=9374",
"bitString=010010010011111 c=> count=5 bits=15 value=9375",
"bitString=0100100101 c=J count=417 bits=10 value=293",
"bitString=01001001100 c=5 count=209 bits=11 value=588",
"bitString=01001001101 c=: count=203 bits=11 value=589",
"bitString=0100100111 c=2 count=410 bits=10 value=295",
"bitString=01001010 c=1 count=1583 bits=8 value=74",
"bitString=01001011 c=0 count=1373 bits=8 value=75",
"bitString=01001100 c=X count=1291 bits=8 value=76",
"bitString=0100110100 c=Q count=398 bits=10 value=308",
"bitString=01001101010 c=3 count=170 bits=11 value=618",
"bitString=01001101011 c=4 count=168 bits=11 value=619",
"bitString=010011011000 c=9 count=88 bits=12 value=1240",
"bitString=010011011001 c=@ count=69 bits=12 value=1241",
"bitString=01001101101 c=8 count=124 bits=11 value=621",
"bitString=0100110111 c=Z count=257 bits=10 value=311",
"bitString=0100111 c=V count=2463 bits=7 value=39",
"bitString=0101 c=O count=22979 bits=4 value=5",
"bitString=0110 c=A count=22752 bits=4 value=6",
"bitString=0111 c=I count=22621 bits=4 value=7",
"bitString=10000 c=L count=11192 bits=5 value=16",
"bitString=10001 c=C count=10543 bits=5 value=17",
"bitString=1001 c=S count=20363 bits=4 value=9",
"bitString=1010 c=N count=20027 bits=4 value=10",
"bitString=10110 c=D count=9794 bits=5 value=22",
"bitString=101110 c=Y count=4749 bits=6 value=46",
"bitString=101111 c=W count=4499 bits=6 value=47",
"bitString=110 c=E count=38211 bits=3 value=6",
"bitString=1110 c=R count=18982 bits=4 value=14",
"bitString=11110 c=M count=8760 bits=5 value=30",
"bitString=11111 c=U count=8053 bits=5 value=31"};
*/

 String[] basicEnglish={"bitString=0010 c=T count=30539 bits=4 value=2",
"bitString=00110 c=H count=14892 bits=5 value=6",
"bitString=001110 c=F count=7763 bits=6 value=14",
"bitString=001111 c=P count=6977 bits=6 value=15",
"bitString=000 c=  count=59159 bits=3 value=0",
"bitString=010000 c=B count=6642 bits=6 value=16",
"bitString=010001 c=G count=6527 bits=6 value=17",
"bitString=01001000 c=K count=1841 bits=8 value=72",
"bitString=010010010000 c=6 count=123 bits=12 value=1168",
"bitString=010010010001 c=? count=123 bits=12 value=1169",
"bitString=010010010010 c== count=118 bits=12 value=1170",
"bitString=0100100100110 c=7 count=68 bits=13 value=2342",
"bitString=01001001001110 c=; count=26 bits=14 value=4686",
"bitString=010010010011110 c=< count=12 bits=15 value=9374",
"bitString=010010010011111 c=> count=5 bits=15 value=9375",
"bitString=0100100101 c=J count=417 bits=10 value=293",
"bitString=01001001100 c=5 count=209 bits=11 value=588",
"bitString=01001001101 c=: count=203 bits=11 value=589",
"bitString=0100100111 c=2 count=410 bits=10 value=295",
"bitString=01001010 c=1 count=1583 bits=8 value=74",
"bitString=01001011 c=0 count=1373 bits=8 value=75",
"bitString=01001100 c=X count=1291 bits=8 value=76",
"bitString=0100110100 c=Q count=398 bits=10 value=308",
"bitString=01001101010 c=3 count=170 bits=11 value=618",
"bitString=01001101011 c=4 count=168 bits=11 value=619",
"bitString=010011011000 c=9 count=88 bits=12 value=1240",
"bitString=010011011001 c=@ count=69 bits=12 value=1241",
"bitString=01001101101 c=8 count=124 bits=11 value=621",
"bitString=0100110111 c=Z count=257 bits=10 value=311",
"bitString=0100111 c=V count=2463 bits=7 value=39",
"bitString=0101 c=O count=22979 bits=4 value=5",
"bitString=0110 c=A count=22752 bits=4 value=6",
"bitString=0111 c=I count=22621 bits=4 value=7",
"bitString=10000 c=L count=11192 bits=5 value=16",
"bitString=10001 c=C count=10543 bits=5 value=17",
"bitString=1001 c=S count=20363 bits=4 value=9",
"bitString=1010 c=N count=20027 bits=4 value=10",
"bitString=10110 c=D count=9794 bits=5 value=22",
"bitString=101110 c=Y count=4749 bits=6 value=46",
"bitString=101111 c=W count=4499 bits=6 value=47",
"bitString=110 c=E count=38211 bits=3 value=6",
"bitString=1110 c=R count=18982 bits=4 value=14",
"bitString=11110 c=M count=8760 bits=5 value=30",
"bitString=11111 c=U count=8053 bits=5 value=31"};
 
   StringConverterNode[] charArray=new StringConverterNode[256];
   
   
   public String fileName="";
   
   public void setFileName(String s){
     fileName=s;
     Vector v=fileToData(s);
     vectorToArray(v);// this is vestigial right now.
   }
 
   public void useBasicEnglish(){
     fileName="";
     Vector v=new Vector();
     for (int i=0;i<basicEnglish.length;i++){
       String l=basicEnglish[i];
       v.addElement(pullApartLine(l));
     }
     vectorToArray(v);
   
   }
 
   public Vector fileToData(String a){
    Vector ans=new Vector();
    
    FileInputStream fis;
    DataInputStream dis;
    String line;
    try{
      fis = new FileInputStream(a);
      dis=new DataInputStream(fis);

      while ((line=dis.readLine())!=null) {
         if (line.length()>0){
             ans.addElement(pullApartLine(line));
            }
      }
      fis.close();
    } catch (FileNotFoundException e){
          System.err.println("Can't find the file."+e);
    } catch (IOException e){
      System.err.println("IO Exception."+e);
    } catch (Exception e){
      System.err.println("General Exception."+e);
    }
    return ans;
   }
	

    public StringConverterNode pullApartLine(String s){
      int i=s.indexOf("c=")+2;
      int i2=s.indexOf(' ',i);
      StringConverterNode ans=new StringConverterNode();
      if (i==i2){
        ans.c=" ";
      } else { 
        ans.c=s.substring(i,i2);
      }
      try{    
	    i=s.indexOf("bits=")+5;
        i2=s.indexOf(' ',i);
        ans.bits=Integer.parseInt(s.substring(i,i2));
      
	    i=s.indexOf("value=")+6;
        i2=s.length(); // value is the last one.
        
        ans.value=Integer.parseInt(s.substring(i,i2));
      } catch (Exception e){
      
      }     
 	    i=s.indexOf("bitString=")+10;
        i2=s.indexOf(' ',i);
        ans.bitString= s.substring(i,i2);
    
      charArray[(byte)ans.c.charAt(0)]=ans;    
      return ans;
    }

    
    StringConverterNode[] ar;
    
    
    public void vectorToArray(Vector v){
      int sz=v.size();
      ar=new StringConverterNode[sz];
      for (int i=0;i<sz;i++){
        ar[i]=(StringConverterNode)v.elementAt(i);
      }
    
    }
    
     public StringConverterNode findNode(char c){
	  return charArray[(byte)c];
	}

	
	public StringConverterNode findNode(String s){
	   int sz=ar.length;
	   for (int i=0;i<sz;i++){
	     if (s.equals(ar[i].c)){
	       return ar[i];
	     }
	   }
	  return null;
	}

    public StringConverterNode findNodeByBitString(String s){
	   int sz=ar.length;
	   for (int i=0;i<sz;i++){
	     if (s.equals(ar[i].bitString)){
	       System.out.println("Decoding:"+s+" to get:"+ar[i].c);
	       return ar[i];
	     }
	   }
	  return null;
	}

    public String englishStringToBitString(String s){
      int sz=s.length();
      StringBuffer ans=new StringBuffer();
      StringConverterNode sn;
      for (int i=0;i<sz;i++){
        sn=findNode(s.charAt(i));// sigh, this is probably slow.
        if (sn!=null){
       //   System.out.println("Converting c="+s.charAt(i)+" to get:"+sn.bitString);
          ans.append(sn.bitString);
        }
      }
    //  System.out.println("Total string is:"+ans.toString());
      return ans.toString();
    }
    
     long[] powersOfTwo={1,2,4,8,16,32,64,128,256,2*256,4*256,8*256,16*256,32*256,64*256,128*256,65536,65536*2,65536*4,65536*8,65536*16,65536*32,65536*64,65536*128,65536*256,65536*256*2,65536*256*4,65536*256*8,65536*256*16,65536*256*32,65536*256*64,65536*256*128};
     
    
     public long englishStringToLong(String s){
      int sz=s.length();
      long ans=1;// It must start with a one-bit
      StringConverterNode sn;
      for (int i=0;i<sz;i++){
        sn=findNode(s.charAt(i));
        if (sn!=null){
          ans=ans*powersOfTwo[sn.bits]+sn.value;
      //    System.out.println("Converting c="+s.charAt(i)+" to get:"+sn.bitString);

        }
      }
      return ans;
    }
     
    
    public BigInteger englishStringToBigInteger(String s){
      int sz=s.length();
      BigInteger ans=BigInteger.valueOf(1); // it must start with a one-bit.
      StringConverterNode sn;
      for (int i=0;i<sz;i++){
        sn=findNode(s.charAt(i));
        if (sn!=null){
          ans=ans.multiply(BigInteger.valueOf(powersOfTwo[sn.bits])).add(BigInteger.valueOf(sn.value));
      //          System.out.println("After round "+i+" to:"+ans.toString(2)+" and adding:"+BigInteger.valueOf(sn.value).toString(2));

        }
      }
    //  System.out.println("Converted to:"+ans.toString(2));
      return ans;
    }
     
    public String bigIntegerToEnglishString(BigInteger b){
      StringBuffer ans=new StringBuffer();
      String bits=b.toString(2);
      System.out.println("Converting="+bits);
      int lastPos=1; // ignore the first one. It must start with a one!
      int pos=2;
      int len=bits.length();
      String temp;
      StringConverterNode sn;
      while (pos<len){
        temp=bits.substring(lastPos,pos);
        sn=findNodeByBitString(temp);
        if (sn!=null){
          ans.append(sn.c);
          lastPos=pos;
          pos=lastPos+1;
        } else {
          pos++;
        }      
      }
       temp=bits.substring(lastPos,pos);
        sn=findNodeByBitString(temp);
        if (sn!=null){
          ans.append(sn.c);
           
        }       
      return ans.toString();
      
    }


    public BigInteger appendLong(BigInteger ans, long value, int bits){
       return ans.multiply(BigInteger.valueOf(powersOfTwo[bits])).add(BigInteger.valueOf(value));   
    }


}